//***************************************************************************//
//*                                                                         *//
//* Unit:    xml_parameters                                                 *//
//* Author:  Dushkin R.V.                                                   *//
//* Date:    27.04.01                                                       *//
//*                                                                         *//
//*  This module contains the class, which describes "Node" object of XML   *//
//* document.                                                               *//
//*                                                                         *//
//***************************************************************************//
unit xml_node;

//-[ INTERFACE ]---------------------------------------------------------------
interface

uses
  classes,
  xml_attribute,
  xml_functions,
  xml_types;

type TXMLNode = class (TObject)
  private
    fTagName    : String;
    fTextData   : String;
    fLevel      : Integer;
    fTagType    : TTagType;
    fChildren   : TList;
    fAttributes : TList;

  public
    constructor Create;
    destructor Destroy; override;

    procedure SetAttributeValue (const attrName, attrValue : String);
    procedure SaveToFile (var where : TextFile);
    procedure SetLevel (iLevel : Integer);
    procedure DeleteAttribute (const attrName : String);
    procedure DeleteChild (iNumber : Integer);
    procedure DeleteChildren;

    function ParseString (iLevel : Integer; var iStr : String) : TXMLErrorCode;
    function ParseAttributes (var iStr : String) : TXMLErrorCode;
    function ParseChildren (var iStr : String) : TXMLErrorCode;
    function FindTagByName (const tagName : String) : TXMLNode;
    function GetAttributeValue (const attrName : String) : String;
    function GetChild (iNumber : Integer) : TXMLNode;
    function GetChildNumber (iChild : TXMLNode) : Integer;
    function GetNumberOfChildren : Integer;
    function AddChild (const tagName : String; tagType : TTagType) : TXMLNode;
    function AddChildAsObject (Child : TXMLNode) : TXMLNode;

  published
    property TagName    : String read fTagName write fTagName;
    property TextData   : String read fTextData write fTextData;
    property Level      : Integer read fLevel write SetLevel;
    property TagType    : TTagType read fTagType write fTagType;
    property Children   : TList read fChildren;
    property Attributes : TList read fAttributes;

end;

//-[ IMPLEMENTATION ]----------------------------------------------------------
implementation

constructor TXMLNode.Create;
begin
  inherited Create;
  fTagName := '';
  fTextData := '';
  fLevel := 0;
  fTagType := ttNormal;
  fChildren := TList.Create;
  fAttributes := TList.Create;
end;

{ DONE 1 -cMethod : Implement good destructor for XMLNode, which frees memory for all children and attributes. }
destructor TXMLNode.Destroy;
var
  i         : Integer;
  child     : TXMLNode;
  attribute : TXMLAttribute;
begin
  for i := 0 to fChildren.Count - 1 do
  begin
    child := fChildren.Items[i];
    child.Free;
  end;
  fChildren.Clear;
  fChildren.Free;
  for i := 0 to fAttributes.Count - 1 do
  begin
    attribute := fAttributes.Items[i];
    attribute.Free;
  end;
  fAttributes.Clear;
  fAttributes.Free;
  inherited Destroy;
end;

function TXMLNode.ParseString (iLevel : Integer; var iStr : String) : TXMLErrorCode;
var
  i            : Integer;
  tempStr      : String;
  noAttributes : Boolean;
begin
  fLevel := iLevel;
  Result := ecOk;
  noAttributes := False;
  if (iStr[1] = '<') then
  begin
    i := 2;
    fTagName := '';
    while ((iStr[i] <> ' ') and (iStr[i] <> '/') and (iStr[i] <> '>')) do
    begin
      fTagName := fTagName + iStr[i];
      i := i + 1;
    end;
    case (iStr[i]) of
      '/':
        begin
          i := i + 1;
          fTagType := ttShorted;
          noAttributes := True;
        end;
      '>':
        begin
          noAttributes := True;
        end;
    end;
    if (fTagName <> '') then
    begin
      Delete (iStr, 1, i);
      if (not noAttributes) then
      begin
        Result := ParseAttributes (iStr);
      end;
      if (Result = ecOk) then
      begin

        { DONE 1 -cMethod : Here is the ground error. When the tag has as a child the same tag, this error occurs. }
        if (fTagType = ttNormal) then
        begin
          tempStr := '</' + fTagName + '>';
          while (Copy (iStr, 1, Length (tempStr)) <> tempStr) do
          begin
            Result := ParseChildren (iStr);
          end;
          if (Result = ecOk) then
          begin

            Delete (iStr, 1, Length (tempStr));
          end;
        end;

      end;
    end
    else
    begin
      Result := ecInvalidTag;
    end;
  end
  else
  begin
    fTagType := ttCDATA;
    fTextData := '';
    i := 1;
    while ((iStr[i] <> '<') and (i <= Length (iStr))) do
    begin
      fTextData := fTextData + iStr[i];
      i := i + 1;
    end;
    Delete (iStr, 1, i - 1);
  end;
end;

function TXMLNode.ParseAttributes (var iStr : String) : TXMLErrorCode;
var
  i        : Integer;
  quotMark : Integer;
  tempStr  : String;
  nextWord : String;
  strLen   : Integer;
  tempAttr : TXMLAttribute;
begin
  fAttributes.Clear;
  Result := ecOk;
  tempStr := Copy (iStr, 1, Pos ('>', iStr));
  strLen := Length (tempStr);
  Delete (iStr, 1, strLen);
  if (tempStr[strLen - 1] = '/') then
  begin
    fTagType := ttShorted;
  end;
  tempStr := tempStr + ' ';
  while (Pos (' ', tempStr) > 0) do
  begin
    i := 1;
    quotMark := 0;
    nextWord := '';
    while ((quotMark < 2) and (i <= Length (tempStr))) do
    begin
      nextWord := nextWord + tempStr[i];
      if (tempStr[i] = '"') then
      begin
        quotMark := quotMark + 1;
      end;
      i := i + 1;
    end;
    if (i <= Length (tempStr)) then
    begin
      nextWord := nextWord + ' ';
    end;

    Delete (tempStr, 1, Length (nextWord));
    if ((nextWord <> '/> ') and (nextWord <> '> ') and (nextWord <> ' ')) then
    begin
      tempAttr := TXMLAttribute.Create;
      Result := tempAttr.ParseString (nextWord);
      fAttributes.Add (tempAttr);
    end;
  end;
end;

function TXMLNode.ParseChildren (var iStr : String) : TXMLErrorCode;
var
  tempNode : TXMLNode;
begin
  tempNode := TXMLNode.Create;
  Result := tempNode.ParseString (fLevel + 1, iStr);
  fChildren.Add (tempNode);
end;

function TXMLNode.FindTagByName (const tagName : String) : TXMLNode;
var
  i        : Integer;
  tempNode : TXMLNode;
begin
  Result := nil;
  fChildren.Pack;
  for i := 0 to fChildren.Count - 1 do
  begin
    tempNode := fChildren.Items[i];
    if (tempNode.TagName = tagName) then
    begin
      Result := fChildren.Items[i];
    end;
  end;

  if (Result = nil) then
  begin
    for i := 0 to fChildren.Count - 1 do
    begin
      tempNode := fChildren.Items[i];
      Result := tempNode.FindTagByName (tagName);
    end;
  end;
end;

function TXMLNode.GetAttributeValue (const attrName : String) : String;
var
  i        : Integer;
  tempAttr : TXMLAttribute;
begin
  Result := 'undefined';
  fAttributes.Pack;
  for i := 0 to fAttributes.Count - 1 do
  begin
    tempAttr := fAttributes.Items[i];
    if (tempAttr.Name = attrName) then
    begin
      Result := tempAttr.Value;
      Break;
    end;
  end;
end;

procedure TXMLNode.SetAttributeValue (const attrName, attrValue : String);
var
  i        : Integer;
  setFlag  : Boolean;
  tempAttr : TXMLAttribute;
begin
  setFlag := False;
  fAttributes.Pack;
  for i := 0 to fAttributes.Count - 1 do
  begin
    tempAttr := fAttributes.Items[i];
    if (tempAttr.Name = attrName) then
    begin
      tempAttr.Value := attrValue;
      setFlag := True;
      Break;
    end;
  end;
  if (not setFlag) then
  begin
    tempAttr := TXMLAttribute.Create;
    tempAttr.Name := attrName;
    tempAttr.Value := attrValue;
    fAttributes.Add (tempAttr);
  end;
end;

procedure TXMLNode.SaveToFile (var where : TextFile);
var
  i        : Integer;
  tempStr  : String;
  tempAttr : TXMLAttribute;
  tempNode : TXMLNode;
begin
  tempStr := '';
  if (fTagType = ttCDATA) then
  begin
    tempStr := tempStr + TrimSpaces (fTextData);
    Write (where, tempStr);
  end
  else
  begin
    for i := 1 to fLevel do
    begin
      tempStr := tempStr + ' ' + ' ';
    end;
    tempStr := tempStr + '<' + fTagName;
    fAttributes.Pack;
    for i := 0 to fAttributes.Count - 1 do
    begin
      tempAttr := fAttributes.Items[i];
      tempStr := tempStr + ' ' + tempAttr.Name + '="' + tempAttr.Value + '"';
    end;
    if (fTagType = ttShorted) then
    begin
      tempStr := tempStr + ' />';
      WriteLn (where, tempStr);
    end
    else
    begin
      tempStr := tempStr + '>';
      fChildren.Pack;
      if ((GetNumberOfChildren = 1) and (GetChild (0).fTagType = ttCDATA)) then
      begin
        Write (where, tempStr);
      end
      else
      begin
        WriteLn (where, tempStr);
      end;
      for i := 0 to fChildren.Count - 1 do
      begin
        tempNode := fChildren.Items[i];
        tempNode.SaveToFile (where);
      end;
      tempStr := '';
      if ((GetNumberOfChildren <> 1) or (GetChild (0).fTagType <> ttCDATA)) then
      begin
        for i := 1 to fLevel do
        begin
          tempStr := tempStr + ' ' + ' ';
        end;
      end;
      tempStr := tempStr + '</' + fTagName + '>';
      WriteLn (where, tempStr);
    end;
  end;
end;

function TXMLNode.GetChild (iNumber : Integer) : TXMLNode;
begin
  Result := nil;
  if (iNumber <= fChildren.Count - 1) then
  begin
    Result := fChildren.Items[iNumber];
  end;
end;

function TXMLNode.GetChildNumber (iChild : TXMLNode) : Integer;
var
  i        : Integer;
  tempNode : TXMLNode;
begin
  Result := -1;
  for i := 0 to fChildren.Count - 1 do
  begin
    tempNode := fChildren.Items[i];
    if (iChild = tempNode) then
    begin
      Result := i;
      Break;
    end;
  end;
end;

function TXMLNode.GetNumberOfChildren : Integer;
begin
  Result := Children.Count;
end;

function TXMLNode.AddChild (const tagName : String; tagType : TTagType) : TXMLNode;
var
  tempNode : TXMLNode;
begin
  tempNode := TXMLNode.Create;
  tempNode.TagName := tagName;
  tempNode.Level := fLevel + 1;
  tempNode.TagType := tagType;
  fChildren.Add (tempNode);
  Result := tempNode;
end;

function TXMLNode.AddChildAsObject (Child : TXMLNode) : TXMLNode;
begin
  fChildren.Add (Child);
  Result := child;
end;

procedure TXMLNode.SetLevel (iLevel : Integer);
var
  i     : Integer;
  Child : TXMLNode;
begin
  fLevel := iLevel;
  for i := 0 to fChildren.Count - 1 do
  begin
    Child := fChildren.Items[i];
    Child.Level := iLevel + 1;
  end;
end;

procedure TXMLNode.DeleteAttribute (const attrName : String);
var
  i             : Integer;
  tempAttribute : TXMLAttribute;
begin
  for i := 0 to fAttributes.Count - 1 do
  begin
    tempAttribute := fAttributes.Items[i];
    if (attrName = tempAttribute.Name) then
    begin
      fAttributes.Delete (i);
      fAttributes.Pack;
      tempAttribute.Free;
      Exit;
    end;
  end;
end;

procedure TXMLNode.DeleteChild (iNumber : Integer);
var
  tempNode : TXMLNode;
begin
  if (iNumber <= fChildren.Count) then
  begin
    tempNode := fChildren.Items[iNumber];
    tempNode.Free;
    fChildren.Delete (iNumber);
    fChildren.Pack;
  end;
end;

procedure TXMLNode.DeleteChildren;
var
  i         : Integer;
  childNode : TXMLNode;
begin
  for i := 0 to fChildren.Count - 1 do
  begin
    childNode := fChildren.Items[i];
    childNode.Free;
  end;
  fChildren.Clear;
  fChildren.Pack;
end;

//-[ END ]---------------------------------------------------------------------
end.
